#define AT_NORTHEAST  16
#define AT_SOUTHEAST  32
#define AT_SOUTHWEST  64
#define AT_NORTHWEST 128

#define at256_reduce47(at) ((at)&15) + (((at)&21)==21 ? 16 : 0) + (((at)&38)==38 ? 32 : 0) + (((at)&74)==74 ? 64 : 0) + (((at)&137)==137 ? 128 : 0)
#define at256_reduce15(at) (at&15)

var
	/* autotile rules: format: list(AUTOTILE_RULE_NAME = list(TILE_ID_NAME=1))
		autotile rules can contain multiple tile ids to join with. */
	//list/autotile_rules = list()
	list/__atlib_dirs = list(NORTH,NORTHEAST,EAST,SOUTHEAST,SOUTH,SOUTHWEST,WEST,NORTHWEST) //this is just used for quick iteration through an 8-dir scheme.
	list/__atlib_dirs_opp = list(SOUTH,AT_SOUTHWEST,WEST,AT_NORTHWEST,NORTH,AT_NORTHEAST,EAST,AT_SOUTHEAST) //this is just used for a quick association to dir16 directions.

turf
	var
		join_id = null //Used for joining rules.
		autotile_icon = null
		autotile_rule = null //the autotile rule group that this tile belongs to. Leave null if this isn't an autotile.
		autotile_reduce = 47
		autotile_flags = 0 //the current state of neighboring turfs in AT256 format.
	proc
		/* args: Dir - the autotile direction that the update is facing toward. Only supply if this is a secondary update from a neighbor instantiation.
		         tile - the neighboring tile that initiated this update. Only supply if this is a secondary update.

		   returns: nothing

		   action: This will update the autotile state of this tile. If this tile has an autotile rule set, it will attempt to look at neighboring tiles
		   		  and generate an AT256 value from the 8 neighboring tiles. It will also call secondary updates on the neighboring tiles.
		   		  If this is a secondary update, no additional updates on neighbors are triggered. Only the corresponding neighbor bit is turned on or off
		   		  based on the autotile rule of this tile matched with the neighbor's tile id.*/
		UpdateAutotile(Dir,turf/tile)
			//approximate speed: 35,000 tiles per second @ 3.6ghz
			if(initialized) //if map_loading, autotile uses an optimized algorithm inlined under turf/New() to reduce loading times for large worlds
				if(autotile_rule)
					if(Dir) //usually called from a neighboring turf update
						if(!tile||(tile.join_id&&autotile_rules[autotile_rule][tile.join_id]))
							autotile_flags |= Dir
						else
							autotile_flags &= ~Dir
					else //usually called on turf/new()
						var/atdir = 0
						var/list/l = autotile_rules[autotile_rule]
						var/turf/t = get_step(src,NORTH)
						if(t) //if this tile has a neighbor (not a map edge)
							if(t.join_id&&l[t.join_id]) atdir |= NORTH //turn on the neighbor bit if the neighbor is in this group.
							if(t.autotile_rule) t.UpdateAutotile(SOUTH,src) //trigger secondary update if the neighbor is an autotile.
						else //otherwise, force a join because this tile is on the edge of the map.
							atdir |= NORTH
						t = get_step(src,EAST)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= EAST
							if(t.autotile_rule) t.UpdateAutotile(WEST,src)
						else
							atdir |= EAST
						t = get_step(src,SOUTH)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= SOUTH
							if(t.autotile_rule) t.UpdateAutotile(NORTH,src)
						else
							atdir |= SOUTH
						t = get_step(src,WEST)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= WEST
							if(t.autotile_rule) t.UpdateAutotile(EAST,src)
						else
							atdir |= WEST
						t = get_step(src,NORTHEAST)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= AT_NORTHEAST
							if(t.autotile_rule) t.UpdateAutotile(AT_SOUTHWEST,src)
						else
							atdir |= AT_NORTHEAST
						t = get_step(src,SOUTHEAST)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= AT_SOUTHEAST
							if(t.autotile_rule) t.UpdateAutotile(AT_NORTHWEST,src)
						else
							atdir |= AT_SOUTHEAST
						t = get_step(src,SOUTHWEST)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= AT_SOUTHWEST
							if(t.autotile_rule) t.UpdateAutotile(AT_NORTHEAST,src)
						else
							atdir |= AT_SOUTHWEST
						t = get_step(src,NORTHWEST)
						if(t)
							if(t.join_id&&l[t.join_id]) atdir |= AT_NORTHWEST
							if(t.autotile_rule) t.UpdateAutotile(AT_SOUTHEAST,src)
						else
							atdir |= AT_NORTHWEST

						autotile_flags = atdir //set the autotile flags
				else if(!Dir) //update neighbors if this is not an autotile
					var/turf/t
					for(var/count in 1 to 8)
						t = get_step(src,__atlib_dirs[count])
						if(t&&t.autotile_rule) t.UpdateAutotile(__atlib_dirs_opp[count],src)
			else
				if(!isnull(icon_state)&&icon_state!="")
					return
			if(autotile_rule)
				switch(autotile_reduce)
					if(256)
						icon_state = "[autotile_flags]" //use raw at256 everywhere
					if(47)
						icon_state = "[at256_reduce47(autotile_flags)]" //distill the at256 value down to at47.
					if(15)
						icon_state = "[at256_reduce15(autotile_flags)]" //distill the at256 value down to at15.

	New()
		if(autotile_icon&&icon!=autotile_icon)
			icon = autotile_icon
			if(autotile_rule&&!isnull(icon_state)&&icon_state!="") icon_state = null
		if(!initialized)
			//this is an optimized version of UpdateAutotile. It only needs to check 4 neighbors rather than 8 because the map initializes from top-right to bottom-left.
			//since all tiles are initialized in order, we can save quite a few instructions by doing this slightly differently.
			//Approximate speed: 80,000 tiles per second @ 3.6ghz
			var/tx = x==1,ty = y==1
			if(autotile_rule)
				var/turf/t = get_step(src,EAST)
				var/rule
				var/list/l = autotile_rules[autotile_rule]
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= WEST
						if(ty) t.UpdateAutotile()
					if(l[t.join_id]) autotile_flags |= EAST
				else
					autotile_flags |= EAST
				t = get_step(src,NORTHEAST)
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= AT_SOUTHWEST
						t.UpdateAutotile()
					if(l[t.join_id]) autotile_flags |= AT_NORTHEAST
				else
					autotile_flags |= AT_NORTHEAST
				t = get_step(src,NORTH)
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= SOUTH
						if(tx)
							t.UpdateAutotile()
					if(l[t.join_id]) autotile_flags |= NORTH
				else
					autotile_flags |= NORTH
				t = get_step(src,NORTHWEST)
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= AT_SOUTHEAST
					if(l[t.join_id]) autotile_flags |= AT_NORTHWEST
				else
					autotile_flags |= AT_NORTHWEST
				if(tx)
					if(ty)
						autotile_flags |= SOUTH|WEST|AT_SOUTHWEST|AT_NORTHWEST|AT_SOUTHEAST
						UpdateAutotile()
					else
						autotile_flags |= WEST|AT_NORTHWEST|AT_SOUTHWEST
				else if(ty)
					autotile_flags |= SOUTH|AT_SOUTHWEST|AT_SOUTHEAST
				else if(x==world.maxx)
					autotile_flags |= AT_SOUTHEAST
			else
				var/turf/t = get_step(src,EAST)
				var/rule
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= WEST
						if(ty) t.UpdateAutotile()
				t = get_step(src,NORTHEAST)
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= AT_SOUTHWEST
						t.UpdateAutotile()
				t = get_step(src,NORTH)
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= SOUTH
						if(tx) t.UpdateAutotile()
				t = get_step(src,NORTHWEST)
				if(t)
					rule = t.autotile_rule
					if(rule)
						if(autotile_rules[rule][join_id]) t.autotile_flags |= AT_SOUTHEAST
		else
			UpdateAutotile()
		..()

#ifdef DEBUG

var
	list/__atlib_states = list(  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
						   21, 23, 29, 31, 38, 39, 46, 47, 55, 63, 74, 75, 78, 79, 95,110,
						  111,127,137,139,141,143,157,159,175,191,203,207,223,239,255)
	                     //  0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  21  23  29  31  38  39  46  47  55  63  74  75  78  79  95 110 111 127 137 139 141 143 157 159 175 191 203 207 223 239 255
	list/__atlib_tl = list(  1,  7,  4,  7,  4,  7,  4,  7,  5,  3,  5,  3,  5,  3,  5,  3, 10,  7,  3,  3,  4,  7,  5,  3,  7,  3,  6,  3,  5,  3,  3,  5,  3,  3, 12,  9, 11,  8, 11,  8,  8,  8,  9,  8,  8,  8,  8)
	list/__atlib_tr = list(  1,  9,  6,  9,  5,  3,  5,  3,  6,  9,  6,  9,  5,  3,  5,  3, 10,  7,  8,  8,  4,  3,  5,  3,  7,  8,  6,  9,  5,  3,  8,  5,  3,  8, 12,  9,  3,  3, 11,  8,  3,  8,  9,  3,  8,  3,  8)
	list/__atlib_bl = list(  1, 10,  7,  7, 10, 10,  7,  7, 11, 11,  3,  3, 11, 11,  3,  3, 10,  7, 11,  3,  4,  7,  3,  3,  7,  3,  6,  8,  5,  8,  8,  5,  8,  8, 12,  3, 11,  3, 11,  3,  3,  3,  9,  8,  8,  8,  8)
	list/__atlib_br = list(  1, 12,  9,  9, 11, 11,  3,  3, 12, 12,  9,  9, 11, 11,  3,  3, 10,  3, 11,  3,  4,  7,  8,  8,  7,  8,  6,  9,  3,  3,  3,  5,  8,  8, 12,  9, 11,  3, 11,  3,  8,  8,  9,  3,  3,  8,  8)


client/verb
	GenAutotile(icon/i as icon)
		set hidden = 1
		set instant = 1
		if(world.host!=key&&world.address)
			return 0
		i = icon(i)
		var/icon/j = new(), icon/k
		var/ix,iy
		var/h = i.Height(), w = i.Width(), ih = floor(h/4), iw = floor(w/3), hw = iw/2, hh = ih/2
		var/count

		while(count<12)
			k = icon(i)
			ix = (count%3)*iw
			iy = h-(floor(count/3)+1)*ih
			k.Crop(ix+1,iy+1,ix+iw,iy+ih)
			j.Insert(k,"[++count]")

		var/icon/bli=new(j),icon/bri=new(j),icon/tli=new(j),icon/tri=new(j)
		bli.Crop(1,1,hw,hh)
		bli.Crop(1,1,iw,ih)
		bri.Crop(hw+1,1,iw,hh)
		bri.Crop(1-hw,1,hw,ih)
		tli.Crop(1,hh+1,hw,ih)
		tli.Crop(1,1-hh,iw,hh)
		tri.Crop(hw+1,hh+1,iw,ih)
		tri.Crop(1-hw,1-hh,hw,hh)

		var/list/bll = list(), list/brl = list(), list/tll = list(), list/trl = list()

		var/lbl
		for(count in 1 to 12)
			lbl = "[count]"
			bll += icon(bli,lbl)
			brl += icon(bri,lbl)
			tll += icon(tli,lbl)
			trl += icon(tri,lbl)

		bli = new(); bri = new(); tli = new(); tri = new()

		for(count in 1 to 47)
			lbl = "[__atlib_states[count]]"
			bli.Insert(bll[__atlib_bl[count]],lbl)
			bri.Insert(brl[__atlib_br[count]],lbl)
			tli.Insert(tll[__atlib_tl[count]],lbl)
			tri.Insert(trl[__atlib_tr[count]],lbl)

		bli.Blend(bri,ICON_OVERLAY)
		tli.Blend(tri,ICON_OVERLAY)
		bli.Blend(tli,ICON_OVERLAY)

		usr << ftp(bli)

	BlendRoof(icon/i as icon,icon/j as icon)
		set hidden = 1
		set instant = 1
		if(world.host!=key&&world.address)
			return 0
		i = icon(i); j = icon(j)
		i.Blend(j,ICON_OVERLAY,1,1+j.Height())
		usr << ftp(i)

#endif